/* Template project which demonstrates the basics on how to setup a project 
 * example application.
 * Device code.
 */
#ifndef _TEMPLATE_KERNEL_H_
#define _TEMPLATE_KERNEL_H_

#include <stdio.h>
#include "cuda.h" 
#include "cutil_math.h" 

#define SDATA( index)      CUT_BANK_CHECKER(sdata, index)


#define arraytexFetch(_texbasename, _tu, _tv, _texnum, _return)\
{\
switch(_texnum)\
{\
case 0: _return = tex2D(_texbasename##0, (_tu), (_tv)); break;\
case 1: _return = tex2D(_texbasename##1, (_tu), (_tv)); break;\
case 2: _return = tex2D(_texbasename##2, (_tu), (_tv)); break;\
case 3: _return = tex2D(_texbasename##3, (_tu), (_tv)); break;\
case 4: _return = tex2D(_texbasename##4, (_tu), (_tv)); break;\
case 5: _return = tex2D(_texbasename##5, (_tu), (_tv)); break;\
case 6: _return = tex2D(_texbasename##6, (_tu), (_tv)); break;\
case 7: _return = tex2D(_texbasename##7, (_tu), (_tv)); break;\
}\
}

texture<unsigned char, 2, cudaReadModeElementType> tex0;
texture<unsigned char, 2, cudaReadModeElementType> tex1;
texture<unsigned char, 2, cudaReadModeElementType> tex2;
texture<unsigned char, 2, cudaReadModeElementType> tex3;
texture<unsigned char, 2, cudaReadModeElementType> tex4;
texture<unsigned char, 2, cudaReadModeElementType> tex5;
texture<unsigned char, 2, cudaReadModeElementType> tex6;
texture<unsigned char, 2, cudaReadModeElementType> tex7;


#define THRESH_CONTOUR_LENGTH 128
#define SCAN_LINE_STRIDE 8  
__constant__ int search_order_table[8][2] = {
{0, -1}, {1, 0},{0, 1}, {-1, 0},
{-1, -1}, {1, -1},{1, 1}, {-1,1}
} ; 



//__device__ void findContour(unsigned int _cv, unsigned int _stx, unsigned int _sty, 
//							                 int2** _contours, unsigned int *contour_pos, 
//											 unsigned char* _m, 
//											 //unsigned char** _maps, 
//											 unsigned int _w, unsigned int _h) 
//{
//	int2** contours = _contours ; 
//	contours += _cv ;
//	int2* _cc = *contours ;

	//unsigned char** maps = _maps ;
	//maps += _cv ;
	//unsigned char* _m = *maps ;

//	_cc[_cv].x =_cv;
//
//	int2* start_pos = _cc ;
//	start_pos += (*contour_pos) ;
//				(*start_pos).x = _cv ;
//				(*start_pos).y = _cv ;
//
//}
//
//
__device__ void findContour(unsigned int _cv, unsigned int _stx, unsigned int _sty, 
							                 int2* _cc, unsigned int *cp, 
											 unsigned char* _m, 
											 unsigned int _w, unsigned int _h) 
{


	int neighbor_x, neighbor_y ;
	int cur_x, cur_y ;
	int next_x, next_y ;
	unsigned int cur_index  ;

	unsigned int order = 0 ;
	unsigned int num_pts = 0 ;
	unsigned int map_index = 0 ;
	unsigned int prev_index = 0 ;

	unsigned char pix_val = 0 ;
	unsigned char pix_val2 = 0;
	int2* start_pos = NULL ;

	bool found = true ;
	bool b_return = false ;
	//bool is_edge = false ;

	cur_x = _stx ;
	cur_y = _sty ;


	while (found == true && !b_return) 							
	{
		found = false ;
		start_pos = NULL ;

#if defined(__DEVICE_EMULATION__)&defined(_DEBUG)
		printf("current_pt: %d %d\n", cur_x, cur_y) ;
#endif
		cur_index = cur_x + cur_y * _w ;
		for(order=0; order<8; order++) 
		{
			
			if(found) 
				break ;

			neighbor_x = (cur_x + search_order_table[order][0])  ;
			neighbor_y = (cur_y + search_order_table[order][1]) ;
						
			// skip pixels outsize the image 
			if(neighbor_x < 0 || neighbor_x > _w-1 || 
				neighbor_y < 0 || neighbor_y > _h-1) 
				continue ; 

			// index of neighbor
			int index = neighbor_x + neighbor_y*_w ;

			// Is this identical to start point ? 
			if(neighbor_x == _stx && neighbor_y == _sty && prev_index != index) 
			{
				start_pos = _cc ;
				start_pos += (*cp) ;
				(*start_pos).x = _stx ;
				(*start_pos).y = _sty ;

				(*cp) ++ ;
				num_pts ++ ;

				b_return = true ;
				break ;
			}

			// Skip pixels of white or already visited. 
			arraytexFetch(tex, neighbor_x, neighbor_y, _cv, pix_val) ;
			if(pix_val == 0 || _m[index]) 
				continue ; 

#if defined(__DEVICE_EMULATION__)&defined(_DEBUG)
			printf("cur_neighbor: %d %d\n", neighbor_x, neighbor_y) ;
			printf("Test edge pt:...") ;
#endif 

			bool is_edge = true;
			for(int order=0; order<8; order++) 
			{
				unsigned int nei_nei_x = neighbor_x + search_order_table[order][0] ; 
				unsigned int nei_nei_y = neighbor_y + search_order_table[order][1] ;

#if defined(__DEVICE_EMULATION__)&defined(_DEBUG)
				printf("(%d %d)", nei_nei_x, nei_nei_y) ;
#endif 
				arraytexFetch(tex, nei_nei_x, nei_nei_y, _cv, pix_val2) ;
				if(pix_val - pix_val2 != 0) 
				{
					is_edge = true ;
					break ;
				}
				else 
					is_edge = false ;
			}

#if defined(__DEVICE_EMULATION__)&defined(_DEBUG)
			printf("\n") ;
#endif 


			if(is_edge) 
			{
				map_index = index ;

				next_x = neighbor_x ;
				next_y = neighbor_y ;
				found = true ;
			}
		}

		if(found && !b_return) 
		{
			// Adding contour pt 
			prev_index = cur_index ;
			cur_x = next_x ;
			cur_y = next_y ;
			_m[map_index] = 255 ;
			
#if defined(__DEVICE_EMULATION__)&defined(_DEBUG)
			printf("Adding contour pt : (%d %d)\n", cur_x, cur_y) ;
#endif 

			start_pos = _cc ;
			start_pos += (*cp) ;
			(*start_pos).x = cur_x ;
			(*start_pos).y = cur_y ;

			(*cp) ++ ;
			num_pts ++ ;

		}
	} 


#if defined(__DEVICE_EMULATION__)&defined(_DEBUG)
	printf("Contour length: %d points\n" , num_pts) ;
#endif 


	// Examine the contour length 
	if(num_pts < THRESH_CONTOUR_LENGTH || !b_return) 
	{
		start_pos = _cc ;
		start_pos += (*cp - num_pts) ;
		memset(start_pos, 0, sizeof(int2)*(*cp-num_pts)) ;
	}
}



__global__ void contourDetect(int2** _conts, unsigned char** _maps,  unsigned int _nv, unsigned int _w, unsigned int _h) 
{
	unsigned int dx = blockIdx.x * blockDim.x + threadIdx.x ;
	unsigned int dy = blockIdx.y * blockDim.y + threadIdx.y ;
	unsigned int th_idx = dx + dy * _nv  ;
	unsigned int cur_idx = 0 ;
	unsigned int cpos = 0 ; // contour position 
	unsigned char cur_val, next_val =0; 

	if(th_idx < _nv) 
	{
		int2* _ccon  = _conts[th_idx] ;
		unsigned char* _m = _maps[th_idx] ;

		unsigned int x, y ;
		for(y=1; y<_h-1; y+=SCAN_LINE_STRIDE) 
		for(x=1; x<_w-1; x++) 
		{
				cur_idx = x + y*_w ;
				arraytexFetch(tex, x, y, th_idx, cur_val) ;

				if(_m[cur_idx] == 255 || cur_val == 0) 
					continue ;

				arraytexFetch(tex, x+1, y, th_idx, next_val) ;

				if(cur_val == 255 && next_val == 0) 
				{
					_m[cur_idx] = 255 ;
					//findContour(th_idx, x, y, _contours, &cpos, _m, _w, _h) ;
					findContour(th_idx, x, y, _ccon, &cpos, _m, _w, _h) ;
					
					_ccon[cpos].x = -100 ;
					_ccon[cpos].y = -100 ;
				}
		}
	}
}







//__global__ void contourDetect2(int2** _contours, unsigned char** _maps,  unsigned int _nv, unsigned int _w, unsigned int _h) 
//{
//	unsigned int dx = blockIdx.x * blockDim.x + threadIdx.x ;
//	unsigned int dy = blockIdx.y * blockDim.y + threadIdx.y ;
//	unsigned int th_idx = dx + dy * _nv  ;
//
//	
//	unsigned int _cv = th_idx ;
//
//
//	if(th_idx < _nv) 
//	{
//		int contour_pos = 0 ; // contour position 
//		unsigned char cur_val, next_val =0; 
//		unsigned int cur_idx = 0 ;
//
//		int2* this_contour  = _contours[th_idx] ;
//		unsigned char* _m = _maps[th_idx] ;
//
//		unsigned int _stx, _sty ;
//		for(_sty=1; _sty<_h-1; _sty+=SCAN_LINE_STRIDE) 
//		for(_stx=1; _stx<_w-1; _stx++) 
//		{
//				cur_idx = _stx + _sty*_w ;
//				arraytexFetch(tex, _stx, _sty, th_idx, cur_val) ;
//
//				if(_m[cur_idx] == 255 || cur_val == 0) 
//					continue ;
//
//				arraytexFetch(tex, _stx+1, _sty, th_idx, next_val) ;
//
//
//
//				//num_pts = 0 ;
//				if(cur_val == 255 && next_val == 0) 
//				{
//					_m[cur_idx] = 255 ;
//					
//					int neighbor_x, neighbor_y ;
//					int cur_x, cur_y ;
//					int next_x=0, next_y=0 ;
//					unsigned int cur_index  ;
//					unsigned int num_pts = 0 ;
//					
//					unsigned int order = 0 ;
//					unsigned int map_index = 0 ;
//					unsigned int prev_index = 0 ;
//
//					unsigned char pix_val = 0 ;
//					unsigned char pix_val2 = 0;
//					int2* start_pos = NULL ;
//
//					bool b_return = false ;
//
//					cur_x = _stx ;
//					cur_y = _sty ;
//
//
//					while (!b_return && (next_x != -1 && next_y!=-1)) 							
//					{
////						found = false ;
//						start_pos = NULL ;
//
//
//				#ifdef __DEVICE_EMULATION__
//						printf("current_pt: %d %d\n", cur_x, cur_y) ;
//				#endif
//						cur_index = cur_x + cur_y * _w ;
//						for(order=0; order<8; order++) 
//						{
//							neighbor_x = (cur_x + search_order_table[order][0])  ;
//							neighbor_y = (cur_y + search_order_table[order][1]) ;
//										
//							// skip pixels outsize the image 
//							if(neighbor_x < 0 || neighbor_x > _w-1 || 
//								neighbor_y < 0 || neighbor_y > _h-1) 
//								continue ; 
//
//							// index of neighbor
//							int index = neighbor_x + neighbor_y*_w ;
//
//
//
//							// Is this identical to start point ? 
//							if(neighbor_x == _stx && neighbor_y == _sty && prev_index != index) 
//							{
//								//start_pos = this_contour ;
//								//start_pos += contour_pos ;
//								//(*start_pos).x = _stx ;
//								//(*start_pos).y = _sty ;
//								this_contour[contour_pos].x = int(_stx) ;
//								this_contour[contour_pos].x = int(_sty) ;
//
//								//contour_pos += 1;
//								//num_pts ++ ;
//
//								b_return = true ;
//								break ;
//							}
//
//							// Skip pixels of white or already visited. 
//							arraytexFetch(tex, neighbor_x, neighbor_y, _cv, pix_val) ;
//							if(pix_val == 0 || _m[index]) 
//								continue ; 
//
//#ifdef __DEVICE_EMULATION__
//							printf("cur_neighbor: %d %d\n", neighbor_x, neighbor_y) ;
//							printf("Test edge pt:...") ;
//#endif 
//							bool is_edge = false ;
//							for(int order=0; order<8; order++) 
//							{
//								next_x = neighbor_x + search_order_table[order][0] ; 
//								next_y = neighbor_y + search_order_table[order][1] ;
//								arraytexFetch(tex, next_x, next_y, _cv, pix_val2) ;
//
//#ifdef __DEVICE_EMULATION__
//								printf("(%d %d)", next_x, next_y) ;
//#endif 
//								if(pix_val - pix_val2 != 0) 
//								{
//									is_edge =  true ;
//									break ;
//								}
//							}
//
//							next_x = next_y = -1 ;
//
//				#ifdef __DEVICE_EMULATION__
//							printf("\n") ;
//				#endif 
//
//							if(is_edge == true) 
//							{
//								map_index = index ;
//								next_x = neighbor_x ;
//								next_y = neighbor_y ;
//
//								break ;
//							}
//						}
//
//						
//						if(next_x != -1 && next_y != -1 && !b_return) 
//						{
//							prev_index = cur_index ;
//							cur_x = next_x ;
//							cur_y = next_y ;
//							_m[map_index] = 255 ;
//							
//				#ifdef __DEVICE_EMULATION__
//							printf("Adding contour pt : (%d %d)\n", cur_x, cur_y) ;
//				#endif 
//
//							start_pos = this_contour ;
//							start_pos += contour_pos ;
//							(*start_pos).x = cur_x ;
//							(*start_pos).y = cur_y ;
//
//							//contour_pos = contour_pos +1  ;
//							//num_pts ++ ;
//
//						}
//					} 
//
//
//				#ifdef __DEVICE_EMULATION__
//					printf("Contour length: %d points\n" , num_pts) ;
//				#endif 
//
//
//				//	// Examine the contour length 
//					if(num_pts < THRESH_CONTOUR_LENGTH || !b_return) 
//					{
//						start_pos = this_contour ;
//						start_pos += (contour_pos - num_pts) ;
//						memset(start_pos, 0, sizeof(int2)*(contour_pos-num_pts)) ;
//					}
//					
//					this_contour[contour_pos].x = -100 ;
//					this_contour[contour_pos].y = -100 ;
//			}
//		}
//	}
//}

#endif // #ifndef _TEMPLATE_KERNEL_H_
